#!/usr/bin/python

from sys    import stdout
from math   import sqrt, exp, log
from random import seed, random

#-----------------------------------------------------------------------

def whtrain (data, eta=0.2):
    # Widrow-Hoff aka Delta rule training
    theta = wx = wy = 0; u = 1  # initialize the parameters
    while u:                    # while there is an error
        u = 0                   # default: no error
        for x,y,c in data:      # traverse the data points
            d = c-1 if wx*x+wy*y >= theta else c-0
            if d != 0:          # if output is wrong
                theta -= eta*d  # update the parameters
                wx    += eta*d*x
                wy    += eta*d*y; u = 1
    return (theta,wx,wy)

#-----------------------------------------------------------------------

def solve (X,b):
    # solve linear equation system X*a = b (3x3 matrix)
    l11 = sqrt(X[0][0])             # compute Cholesky decomposition
    l21 = X[1][0]            /l11   # of the coefficient matrix
    l22 = sqrt(X[1][1] -l21*l21)    # of the linear equation system,
    l31 = X[2][0]            /l11   # then solve the equation system
    l32 = (X[2][1] -l21*l31) /l22   # with two-step backsubstitution
    l33 = sqrt(X[2][2] -l31*l31 -l32*l32)
    z0  =  b[0]                  /l11
    z1  = (b[1] -l21*z0)         /l22
    z2  = (b[2] -l31*z0 -l32*z1) /l33
    a2  =  z2                    /l33
    a1  = (z1           -l32*a2) /l22
    a0  = (z0   -l21*a1 -l31*a2) /l11
    return (a0,a1,a2)               # return the solution

#-----------------------------------------------------------------------

def regress (data, epsilon=0.1):
    # standard linear regression (x,y -> z)
    e = log(epsilon/(1-epsilon)) if epsilon != 0 else 1
    X = [[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]
    b =  [0.0,0.0,0.0]          # init. matrix and right hand side
    for x,y,z in data:          # traverse the data points
        if epsilon != 0: z = e if z > 0 else -e
        else:            z = log(z/(1-z))
        X[0][0] += 1;                                 b[0] +=   z;
        X[1][0] += x; X[1][1] += x*x;                 b[1] += x*z;
        X[2][0] += y; X[2][1] += y*x; X[2][2] += y*y; b[2] += y*z;
    #X[0][1] = X[1][0]; X[0][2] = X[2][0]; X[1][2] = X[2][1]
    return solve(X,b)           # solve equation system

#-----------------------------------------------------------------------

def err_gd (data, wgts=(0,0,0), eta=0.05, steps=10000, thresh=1e-9):
    # gradient descent on sum of squared errors/deviations
    theta,wx,wy = wgts; o = 0
    #print(theta,wx,wy)
    for s in xrange(1,steps+1): # repeat updates
        dx = dy = dt = e = 0.0  # init. changes and error
        for x,y,c in data:      # traverse the data points
            z = 1.0/(1.0+exp(-(theta +wx*x +wy*y)))
            d = c-z; e += d*d   # compute the error
            d *= z *(1-z)       # and the update factor
            # The preceding line is the only difference to
            # a gradient ascent on the likelihood function.
            dt += d; dx += d*x; dy += d*y
        theta += eta*dt         # update the parameters
        wx    += eta*dx
        wy    += eta*dy
        #print(theta,wx,wy)
        if e < thresh or abs(o-e) < thresh: break
        o = e                   # check for termination
    #print('err_gd: %5d (%g)' % (s,abs(o-e)))
    return (theta,wx,wy)        # return final parameters

#-----------------------------------------------------------------------

def err_nr (data, wgts=(0,0,0), steps=10000, thresh=1e-9):
    # Newton-Raphson on sum of squared errors/deviations
    theta,wx,wy = wgts; o = 0   # Newton-Raphson for error function
    #print(theta,wx,wy)
    for s in xrange(1,steps+1): # repeat updates
        e = 0.0                 # init. sum of squared errors
        g =  [0.0,0.0,0.0]      # init. gradient and Hessian matrix
        H = [[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]
        for x,y,c in data:      # traverse the data points
            z = 1.0/(1.0+exp(-(theta +wx*x +wy*y))); r = z*(1-z)
            d = c-z; e += d*d   # compute the error
            d = -2*d*r          # and the update factor
            g[0]    += d;   g[1]    += d*x;   g[2]    += d*y
            e = -2*(c -2*(c+1)*z +3*z*z) *r
            H[0][0] += e        # update gradient and Hessian matrix
            H[1][0] += e*x; H[1][1] += e*x*x
            H[2][0] += e*y; H[2][1] += e*y*x; H[2][2] += e*y*y
        dt,dx,dy = solve(H,g)   # compute H^{-1} * g
        theta -= dt             # and update the parameters
        wx    -= dx
        wy    -= dy
        #print(theta,wx,wy)
        if e < thresh or abs(o-e) < thresh: break
        o = e                   # check for termination
    #print('err_nr: %5d (%g)' % (s,abs(o-e)))
    return (theta,wx,wy)        # return final parameters

#-----------------------------------------------------------------------

def llh_ga (data, wgts=(0,0,0), eta=0.01, steps=10000, thresh=1e-9):
    # gradient ascent on log-likelihood function
    theta,wx,wy = wgts; o = 0
    #print(theta,wx,wy)
    for s in xrange(1,steps+1): # repeat updates
        dx = dy = dt = e = 0.0  # init. changes and error
        for x,y,c in data:      # traverse the data points
            z = 1.0/(1.0+exp(-(theta +wx*x +wy*y)))
            d = c-z; e += d*d   # compute the error
            dt += d; dx += d*x; dy += d*y
        theta += eta*dt         # update the parameters
        wx    += eta*dx
        wy    += eta*dy
        #print(theta,wx,wy)
        if e < thresh or abs(o-e) < thresh: break
        o = e                   # check for termination
    #print('llh_ga: %5d (%g)' % (s,abs(o-e)))
    return (theta,wx,wy)        # return final parameters

#-----------------------------------------------------------------------

def llh_nr (data, wgts=(0,0,0), steps=10000, thresh=1e-9):
    # Newton-Raphson on log-likelihood function
    theta,wx,wy = wgts; o = 0   # Newton-Raphson for max. likelihood
    #print(theta,wx,wy)
    for s in xrange(1,steps+1): # repeat updates
        e = 0.0                 # init. sum of squared errors
        g =  [0.0,0.0,0.0]      # init. gradient and Hessian matrix
        H = [[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]
        for x,y,c in data:      # traverse the data points
            z = 1.0/(1.0+exp(-(theta +wx*x +wy*y))); r = z*(1-z)
            d = c-z; e += d*d   # compute the error
            g[0]    += d;   g[1]    += d*x;   g[2]    += d*y
            H[0][0] += r        # update gradient and Hessian matrix
            H[1][0] += r*x; H[1][1] += r*x*x
            H[2][0] += r*y; H[2][1] += r*y*x; H[2][2] += r*y*y
        dt,dx,dy = solve(H,g)   # compute H^{-1} * g
        theta += dt             # and update the parameters
        wx    += dx
        wy    += dy
        #print(theta,wx,wy)
        if e < thresh or abs(o-e) < thresh: break
        o = e                   # check for termination
    #print('llh_nr: %5d (%g)' % (s,abs(o-e)))
    return (theta,wx,wy)        # return final parameters

#-----------------------------------------------------------------------

def pshow (parms):
    print('% 20.16f % 20.16f % 20.16f' % parms)
    #s = sqrt(sum([p*p for p in parms]))
    #print('% 20.16f % 20.16f % 20.16f' % tuple(p/s for p in parms))

#-----------------------------------------------------------------------

def mklines (wgts, colors=('0.67white','0.85white')):
    theta,wx,wy = wgts
    return [((theta,             wx,wy),2,colors[0]),
            ((theta-log(0.6/0.4),wx,wy),1,colors[1]),
            ((theta-log(0.4/0.6),wx,wy),1,colors[1]),
            ((theta-log(0.7/0.3),wx,wy),1,colors[1]),
            ((theta-log(0.3/0.7),wx,wy),1,colors[1]),
            ((theta-log(0.8/0.2),wx,wy),1,colors[1]),
            ((theta-log(0.2/0.8),wx,wy),1,colors[1]),
            ((theta-log(0.9/0.1),wx,wy),1,colors[1]),
            ((theta-log(0.1/0.9),wx,wy),1,colors[1])]

#-----------------------------------------------------------------------

def fig2d (fid, data, lines=[((6,2,1),'0.7white')]):
    out.write('beginfig(%d);\n' % fid)
    out.write('  path p,q; q := unitsquare scaled 40u;\n')
    #out.write('  fill unitsquare shifted -(2.8u,3.5u) '
    #                +'withcolor red;\n')
    for p,w,s in lines:
        theta,wx,wy = p   # wx*x + wy*y + theta = 0
        p = tuple((-(theta+y*wy)/wx,y) for y in [-1.0,5.0])
        out.write('  p := ((%f,%f)--' % p[0])
        out.write(        '(%f,%f)) scaled 10u;\n' % p[1])
        out.write('  p := (p intersectionpoint q)--')
        out.write(       '(p intersectionpoint reverse q);\n')
        out.write('  draw p withpen penrazor scaled %gpw ' % w)
        out.write(         'withcolor %s;\n' % s)
    out.write('  pickup pencircle scaled pw;\n')
    out.write('  draw q;\n')
    out.write('  for i = 0,1,2,3,4:\n')
    out.write('      draw ((-0.8u,0)--(0,0)) shifted (0,i*10u);\n')
    out.write('      draw ((0,-0.8u)--(0,0)) shifted (i*10u,0);\n')
    out.write('  endfor\n')
    out.write('  label.lft(btex \\footnotesize 0 etex, (-1.4u, 0u));\n')
    out.write('  label.lft(btex \\footnotesize 1 etex, (-1.4u,10u));\n')
    out.write('  label.lft(btex \\footnotesize 2 etex, (-1.4u,20u));\n')
    out.write('  label.lft(btex \\footnotesize 3 etex, (-1.4u,30u));\n')
    out.write('  label.lft(btex \\footnotesize 4 etex, (-1.4u,40u));\n')
    out.write('  label.bot(btex \\footnotesize 0 etex, ( 0u,-1.6u));\n')
    out.write('  label.bot(btex \\footnotesize 1 etex, (10u,-1.6u));\n')
    out.write('  label.bot(btex \\footnotesize 2 etex, (20u,-1.6u));\n')
    out.write('  label.bot(btex \\footnotesize 3 etex, (30u,-1.6u));\n')
    out.write('  label.bot(btex \\footnotesize 4 etex, (40u,-1.6u));\n')
    for d in data: out.write('  dot(%.6f,%.6f,%.6f);\n' % d)
    out.write('endfig;\n\n')

#-----------------------------------------------------------------------

def fig3d (fid, data):
    out.write('beginfig(%d);\n' % fid)
    out.write('  scplot.initx(points);\n')
    for d in data:
        out.write('  scplot.addx(points, %.6f,' % d[2])
        out.write('%.6f,%.6f,%.6f);\n' % d)
    out.write('  scplot.select(points, 1, 2, 3, 0);\n')
    out.write('  scplot.set_defaults("ptmr8r");\n')
    out.write('  scplot.set_units(10mm, 10mm, 40mm);\n')
    out.write('  scplot.set_ranges((0,4), (0,4), (-0.05,1.05));\n')
    out.write('  scplot.set_view(-32, 20);\n')
    out.write('  scplot.mode           := 0;\n')
    out.write('  scplot.expand         := 0;\n')
    out.write('  scplot.colors[0]      := blue;\n')
    out.write('  scplot.colors[1]      := (1,0.8,0);\n')
    out.write('  axes.z.kinked         := true;\n')
    out.write('  axes.x.grid.thickness := 0.4pt;\n')
    out.write('  axes.y.grid.thickness := 0.4pt;\n')
    out.write('  axes.z.grid.thickness := 0.4pt;\n')
    out.write('  axes.x.alabel.offset  := (0,4.8mm);\n')
    out.write('  axes.y.alabel.offset  := (0,4.8mm);\n')
    out.write('  axes.z.alabel.offset  := (0,5mm);\n')
    out.write('  axes.z.alabel.rotate  := -90;\n')
    out.write('  scplot.draw_xticks(0, 1, 2, 3, 4);\n')
    out.write('  scplot.draw_yticks(0, 1, 2, 3, 4);\n')
    out.write('  scplot.draw_zticks(0, 0.2, 0.4, 0.6, 0.8, 1.0);\n')
    out.write('  scplot.draw_xlabel(0, btex $x_1$ etex);\n')
    out.write('  scplot.draw_ylabel(0, btex $x_2$ etex);\n')
    out.write('  scplot.draw_zlabel(0, btex $y$   etex);\n')
    out.write('  scplot.draw_frame;\n')
    out.write('  scplot.draw_data(0, points);\n')
    out.write('endfig;\n\n')

#-----------------------------------------------------------------------

def write (data, fname):
    with open(fname, 'w') as out:
        for d in data:
            out.write('\t'.join([('%.16f' % x) for x in d]) +'\n')

#-----------------------------------------------------------------------

if __name__ == '__main__':
    with open('logcls.mp', 'w') as out:
        out.write('verbatimtex\n')
        out.write('\\documentclass{article}\n')
        out.write('\\begin{document}\n')
        out.write('etex\n\n')
        out.write('prologues := 1;\n\n')
        out.write('input arrows;\n')
        out.write('input axes;\n')
        out.write('input scplot;\n\n')
        out.write('u  := 1mm;\n')
        out.write('pw := 0.5pt;\n\n')
        out.write('labeloffset := 0;\n\n')
        out.write('ahangle     := 30;\n')
        out.write('ahlength    := 2mm;\n')
        out.write('color yellow; yellow = (1,0.8,0);\n')
        out.write('path d;\n\n')
        out.write('vardef dot (expr x,y,c) =\n')
        out.write('  d := fullcircle scaled u shifted (x*10u,y*10u);\n')
        out.write('  fill d withcolor scplot.hue(c,0);\n')
        out.write('  draw d withpen pencircle scaled 0.67pw;\n')
        out.write('enddef;\n\n')

        seed(1)
        dxy = [(3.9*random()+0.05,
                3.9*random()+0.05) for i in range(100)]
        gxy  = [(x,y)
                for x in [0.2,0.6,1.0,1.4,1.8,2.2,2.6,3.0,3.4,3.8]
                for y in [0.2,0.6,1.0,1.4,1.8,2.2,2.6,3.0,3.4,3.8]]
        seed(1)
        dzz = [random() for i in range(len(dxy))]
        act = mklines((-6,2,1))
        fid = 0                 # initialize the figure id/number

        # --- logistic classification, 2 classes
        out.write('beginfig(%d);\n' % fid); fid += 1
        out.write('  vardef logistic (expr x) = '
                 +'(x,1/(1 +mexp(-x*256))) enddef;\n')
        out.write('  path p;\n')
        out.write('  p = (logistic(-4.8)\n')
        out.write('       for i = -4.6 step 0.2 until 4.8:\n')
        out.write('         ..tension2.. logistic(i) endfor\n')
        out.write('       ..tension2.. logistic(4.8))\n')
        out.write('      xscaled 5u yscaled 20u shifted (28u,5u);\n')
        out.write('  fill p--(52u,25u)--(4u,25u)--cycle '
                 +'withcolor (0.25blue+0.75white);\n')
        out.write('  fill p--(52u, 5u)--(4u, 5u)--cycle '
                 +'withcolor (0.25red +0.75white);\n')
        out.write('  draw (4u,25u)--(52u,25u) '
                 +'withpen pencircle scaled 0.3pw;\n')
        out.write('  drawarrow ( 3u, 5u)--(54u, 5u);\n')
        out.write('  drawarrow ( 4u, 4u)--( 4u,28u);\n')
        out.write('  draw      ( 3u,15u)--( 4u,15u);\n')
        out.write('  draw      ( 3u,25u)--( 4u,25u);\n')
        out.write('  draw      (28u, 4u)--(28u, 5u);\n')
        out.write('  label.top(btex \\footnotesize $x$           etex, '
                             +'(53u, 6.5u));\n')
        out.write('  label.lft(btex \\footnotesize $0$           etex, '
                             +'( 2u, 5.0u));\n')
        out.write('  label.lft(btex \\footnotesize $\\frac{1}{2}$ etex, '
                             +'( 2u,15.0u));\n')
        out.write('  label.lft(btex \\footnotesize $1$           etex, '
                             +'( 2u,25.0u));\n')
        out.write('  for i = 8 step 10 until 48: '
                      +'draw (i*u,4u)--(i*u,5u); endfor\n')
        out.write('  label(btex \\footnotesize '
                          +'$x_0\\!-\\!\\frac{4}{a}$ etex, '
                          +'( 8u,2.0u));\n')
        out.write('  label(btex \\footnotesize '
                          +'$x_0\\!-\\!\\frac{2}{a}$ etex, '
                          +'(18u,2.0u));\n')
        out.write('  label(btex \\footnotesize '
                          +'$\phantom{+}x_0\phantom{+}$ etex, '
                          +'(28u,2.0u));\n')
        out.write('  label(btex \\footnotesize '
                          +'$x_0\\!+\\!\\frac{2}{a}$ etex, '
                          +'(38u,2.0u));\n')
        out.write('  label(btex \\footnotesize '
                          +'$x_0\\!+\\!\\frac{4}{a}$ etex, '
                          +'(48u,2.0u));\n')
        out.write('  draw p withpen pencircle scaled 2pw;\n')
        out.write('  label.bot(btex \\footnotesize '
                             +'probability etex, '
                             +'(40u,11u));\n')
        out.write('  label.bot(btex \\footnotesize class 1 etex, '
                             +'(42u,15u));\n')
        out.write('  label.top(btex \\footnotesize '
                             +'probability etex, '
                             +'(16u,19u));\n')
        out.write('  label.top(btex \\footnotesize class 0 etex, '
                             +'(14u,15u));\n')
        out.write('endfig;\n\n')

        # --- logistic function without noise
        data = [(x, y, 1/(1+exp(-(-6.0 +2*x +1*y)))) for x,y in dxy]
        #write(data, 'nonoise.txt')
        #lreg  = regress(data, 0);   #pshow(lreg)
        #gerr  = err_gd(data, lreg); #pshow(gerr)
        #nerr  = err_nr(data, lreg); #pshow(nerr)
        #gllh  = llh_ga(data, lreg); #pshow(gllh)
        #nllh  = llh_nr(data, lreg); #pshow(nllh)
        #lines = act[0:1]
        #lines += [(lreg, 2, 'black'   )]
        #lines += [(gerr, 2, '0.7green')]
        #lines += [(gllh, 2, '')]
        fig2d(fid,data,act);               fid += 1
        fig3d(fid,data);                   fid += 1

        # --- logistic function with noise
        data = []
        for i in range(len(dxy)):
            x,y = dxy[i]
            c = 1/(1+exp(-(2*x+1*y-6.0)))
            o = 0.5*(dzz[i]-0.5)
            while c+o >= 1 or c+o <= 0: o = 0.5*o
            data.append((x,y,c+o))
        #write(data, 'noise.txt');
        lreg  = regress(data, 0);   #pshow(lreg)
        gerr  = err_gd(data, lreg); #pshow(gerr)
        #nerr  = err_nr(data, lreg); #pshow(nerr)
        gllh  = llh_ga(data, lreg); #pshow(gllh)
        #nllh  = llh_nr(data, lreg); #pshow(nllh)
        lines = act[0:1];
        lines += [(lreg, 2, 'black'      )]
        lines += [(gerr, 2, '0.7green'   )]
        lines += [(gllh, 2, '(0.8,0,0.8)')]
        t2lines = mklines(lreg, ('black','0.5white'))
        fig2d(fid,data,act);               fid += 1
        fig3d(fid,data);                   fid += 1
        fig2d(fid,data,act+t2lines);       fid += 1
        fig2d(fid,data,act+lines[1:-1]);   fid += 1
        fig2d(fid,data,act+lines[1:]);     fid += 1

        # --- logistic function on grid without noise
        data = [(x,y,1/(1+exp(-(-6.0 +2*x +1*y)))) for x,y in gxy]
        fig2d(fid,data,act);               fid += 1
        fig3d(fid,data);                   fid += 1

        # --- logistic function on grid with noise
        data = []
        for i in range(len(gxy)):
            x,y = gxy[i]
            c = 1/(1+exp(-(-6.0 +2*x +1*y)))
            o = 0.5*(dzz[i]-0.5)
            while c+o >= 1 or c+o <= 0: o = 0.5*o
            data.append((x,y,c+o))
        lreg  = regress(data, 0);   #pshow(lreg)
        gerr  = err_gd(data, lreg); #pshow(gerr)
        #nerr  = err_nr(data, lreg); #pshow(nerr)
        gllh  = llh_ga(data, lreg); #pshow(gllh)
        #nllh  = llh_nr(data, lreg); #pshow(nllh)
        lines = act[0:1]
        lines += [(lreg, 2, 'black'      )]
        lines += [(gerr, 2, '0.7green'   )]
        lines += [(gllh, 2, '(0.8,0,0.8)')]
        fig2d(fid,data,act);               fid += 1
        fig3d(fid,data);                   fid += 1
        fig2d(fid,data,act+lines[1:2]);    fid += 1
        fig2d(fid,data,act+lines[1:-1]);   fid += 1
        fig2d(fid,data,act+lines[1:]);     fid += 1

        # --- linearly separable, clear gap
        data = []; d = 0.4
        for x,y in dxy:
            z = 1/(1+exp(-(-6.0 +2*x +1*y)))
            c = 1 if z > 0.5 else 0
            if c > 0: x = 3.95-((3.95-x)*(1+0.5*y-d)/(1+0.5*y))
            else:     x = 0.05+((x-0.05)*(3-0.5*y-d)/(3-0.5*y))
            data.append((x,y,c))
        lreg  = regress(data);      #pshow(lreg)
        gerr  = err_gd(data, lreg); #pshow(gerr)
        #nerr  = err_nr(data, lreg); #pshow(nerr)
        gllh  = llh_ga(data, lreg); #pshow(gllh)
        #nllh  = llh_nr(data, lreg); #pshow(nllh)
        lines = act[0:1]
        #lines += [(whtrain(data),        2, 'yellow'     )]
        lines += [(lreg, 2, 'black'      )]
        lines += [(gerr, 2, '0.7green'   )]
        lines += [(gllh, 2, '(0.8,0,0.8)')]
        fig2d(fid,data,[]);                fid += 1
        fig3d(fid,data);                   fid += 1
        fig2d(fid,data,act);               fid += 1
        fig2d(fid,data,act+lines[1:2]);    fid += 1
        fig2d(fid,data,act+lines[1:-1]);   fid += 1
        fig2d(fid,data,act+lines[1:]);     fid += 1
        
        # --- linearly separable, hardly any gap
        data  = [(x, y, 1 if 2*x+y >= 6 else 0) for x,y in dxy]
        lreg  = regress(data);      #pshow(lreg)
        gerr  = err_gd(data, lreg); #pshow(gerr)
        #nerr  = err_nr(data, lreg); #pshow(nerr)
        gllh  = llh_ga(data, lreg); #pshow(gllh)
        #nllh  = llh_nr(data, lreg); #pshow(nllh)
        lines = act[0:1]
        #lines += [(whtrain(data),        2, 'yellow'  )]
        lines += [(lreg, 2, 'black'      )]
        lines += [(gerr, 2, '0.7green'   )]
        lines += [(gllh, 2, '(0.8,0,0.8)')]
        fig2d(fid,data,[]);                fid += 1
        fig3d(fid,data);                   fid += 1
        fig2d(fid,data,act);               fid += 1
        fig2d(fid,data,act+lines[1:2]);    fid += 1
        fig2d(fid,data,act+lines[1:-1]);   fid += 1
        fig2d(fid,data,act+lines[1:]);     fid += 1

        # --- logistic function based label choice
        data = []
        for i in range(len(dxy)):
            x,y = dxy[i]
            c = 1 if dzz[i] <= 1/(1+exp(-(-6.0 +2*x +1*y))) else 0
            data.append((x,y,c))
        lreg  = regress(data);      #pshow(lreg)
        gerr  = err_gd(data, lreg); #pshow(gerr)
        #nerr  = err_nr(data, lreg); #pshow(nerr)
        gllh  = llh_ga(data, lreg); #pshow(gllh)
        #nllh  = llh_nr(data, lreg); #pshow(nllh)
        lines = act[0:1]
        lines += [(lreg, 2, 'black'      )]
        lines += [(gerr, 2, '0.7green'   )]
        lines += [(gllh, 2, '(0.8,0,0.8)')]
        fig2d(fid,data,[]);                fid += 1
        fig3d(fid,data);                   fid += 1
        fig2d(fid,data,act);               fid += 1
        fig2d(fid,data,act+lines[1:2]);    fid += 1
        fig2d(fid,data,act+lines[1:-1]);   fid += 1
        fig2d(fid,data,act+lines[1:]);     fid += 1

        out.write('end.\n')
